/*
** Module   : Operators.java
** Abstract : Dispatcher methods for Java language, server-side database
**            functions which replace Progress comparison and math operators
**
** Copyright (c) 2004-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20060811   @28585 Created initial version. Backing methods for
**                           Progress comparison and math operators executed
**                           within the database server as custom, 
**                           Java-language functions.
** 002 ECF 20060817   @28691 Removed numerous methods which were causing
**                           ambiguity in server-side function resolution.
**                           All binary operator methods which accepted mixed
**                           lists of Integers and Doubles were eliminated. 
**                           Now we have only methods which accept two 
**                           Integers or two Doubles.
** 003 ECF 20060904   @29218 Added support for modulo operator.
** 004 ECF 20070608   @34015 Replaced all Double with BigDecimal. Reflects
**                           new implementations of decimal and MathOps
**                           classes.
** 005 ECF 20071210   @36495 Implemented error handling for CAN-FINDs.
**                           Built-in functions/operators within converted
**                           CAN-FIND statements must handle errors by
**                           returning null instead of throwing an exception. 
**                           These are wrapped in the 
**                           ErrorHandler.checkError() function, which checks
**                           pending error state and returns an appropriate
**                           result.
** 006 ECF 20080313   @37512 Expanded API to provide variants with
**                           exhaustive combinations of numeric parameter
**                           types. This is necessary to enable "manual"
**                           overloading user defined functions.
** 007 OM  20121222          Modulus operations for int64.
**/

package com.goldencode.p2j.persist.pl;

import java.math.*;
import java.sql.Date;
import com.goldencode.p2j.util.*;

/**
 * A collection of static methods which provide the backing functionality for
 * Progress-like comparison and math operators, which are executed within the
 * database server as custom, Java-language functions.  This provides a much
 * faster alternative to client-side where clause processing (see {@link
 * com.goldencode.p2j.persist.WhereExpression WhereExpression} and its
 * related classes).  This mechanism (or client-side where clause processing)
 * must be used in cases where an operation converted from Progress code must
 * retain semantics specific to the Progress environment.  Executing these
 * operations on the server is much preferred, since moving records across the
 * network only to test them for inclusion in a result set is prohibitively
 * expensive.
 * <p>
 * This class serves primarily as a consolidation point and dispatcher for
 * worker methods located in diverse classes.  All method calls made by the
 * database server are redirected to those worker methods.  However, since
 * the database server knows nothing of the P2J wrapper types used by those
 * methods to represent primitives, this class also performs a transformation
 * to (on input) and from (on output) the P2J wrapper types.  Thus, all
 * parameters and return types for methods in this class are specified in
 * terms of the standard J2SE wrapper types or primitive types.
 * <p>
 * The transformation handles conversion between <code>null</code> J2SE
 * wrapper types and the implementation of <code>unknown value</code> in the
 * P2J wrapper types.  This is possible because unknown value maps to SQL
 * NULL in the backing database.  However, certain data can be lost in the
 * transformation, and if these data are necessary, this mechanism should be
 * avoided.  Instead, client-side where clause processing should be used in
 * these cases.  Specifically, the following data are not represented by J2SE
 * wrappers or primitives and thus cannot be preserved during transformation:
 * <ul>
 *   <li>non-default precision of a <code>decimal</code> value;
 *   <li>non-default case sensitivity of a <code>character</code> value;
 *   <li>non-default timezone of a <code>date</code> value.
 * </ul>
 *
 * @see  Functions
 */
public final class Operators
{
   /**
    * Test two values for equality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean eq(Boolean a, Boolean b)
   {
      try
      {
         return CompareOps.isEqual(new logical(a),
                                   new logical(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test two values for equality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean eq(String a, String b)
   {
      try
      {
         return CompareOps.isEqual(new character(a),
                                   new character(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test two values for equality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean eq(Integer a, Integer b)
   {
      try
      {
         return CompareOps.isEqual(new integer(a),
                                   new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test two values for equality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean eq(BigDecimal a, BigDecimal b)
   {
      try
      {
         return CompareOps.isEqual(new decimal(a),
                                   new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test two values for equality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean eq(BigDecimal a, Integer b)
   {
      try
      {
         return CompareOps.isEqual(new decimal(a),
                                   new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test two values for equality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean eq(Integer a, BigDecimal b)
   {
      try
      {
         return CompareOps.isEqual(new integer(a),
                                   new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test two values for inequality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isNotEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean ne(Boolean a, Boolean b)
   {
      try
      {
         return CompareOps.isNotEqual(new logical(a),
                                      new logical(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test two values for inequality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isNotEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean ne(String a, String b)
   {
      try
      {
         return CompareOps.isNotEqual(new character(a),
                                      new character(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test two values for inequality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isNotEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean ne(Integer a, Integer b)
   {
      try
      {
         return CompareOps.isNotEqual(new integer(a),
                                      new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test two values for inequality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isNotEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean ne(BigDecimal a, BigDecimal b)
   {
      try
      {
         return CompareOps.isNotEqual(new decimal(a),
                                      new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test two values for inequality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isNotEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean ne(BigDecimal a, Integer b)
   {
      try
      {
         return CompareOps.isNotEqual(new decimal(a),
                                      new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test two values for inequality.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isNotEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean ne(Integer a, BigDecimal b)
   {
      try
      {
         return CompareOps.isNotEqual(new integer(a),
                                      new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gt(Boolean a, Boolean b)
   {
      try
      {
         return CompareOps.isGreaterThan(new logical(a),
                                         new logical(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gt(String a, String b)
   {
      try
      {
         return CompareOps.isGreaterThan(new character(a),
                                         new character(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gt(Integer a, Integer b)
   {
      try
      {
         return CompareOps.isGreaterThan(new integer(a),
                                         new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gt(BigDecimal a, BigDecimal b)
   {
      try
      {
         return CompareOps.isGreaterThan(new decimal(a),
                                         new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gt(BigDecimal a, Integer b)
   {
      try
      {
         return CompareOps.isGreaterThan(new decimal(a),
                                         new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gt(Integer a, BigDecimal b)
   {
      try
      {
         return CompareOps.isGreaterThan(new integer(a),
                                         new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gte(Boolean a, Boolean b)
   {
      try
      {
         return CompareOps.isGreaterThanOrEqual(new logical(a),
                                                new logical(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gte(String a, String b)
   {
      try
      {
         return CompareOps.isGreaterThanOrEqual(
            new character(a),
            new character(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gte(Integer a, Integer b)
   {
      try
      {
         return CompareOps.isGreaterThanOrEqual(new integer(a),
                                                new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gte(BigDecimal a, BigDecimal b)
   {
      try
      {
         return CompareOps.isGreaterThanOrEqual(new decimal(a),
                                                new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gte(BigDecimal a, Integer b)
   {
      try
      {
         return CompareOps.isGreaterThanOrEqual(new decimal(a),
                                                new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &gt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isGreaterThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean gte(Integer a, BigDecimal b)
   {
      try
      {
         return CompareOps.isGreaterThanOrEqual(new integer(a),
                                                new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lt(Boolean a, Boolean b)
   {
      try
      {
         return CompareOps.isLessThan(new logical(a),
                                      new logical(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lt(String a, String b)
   {
      try
      {
         return CompareOps.isLessThan(new character(a),
                                      new character(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lt(Integer a, Integer b)
   {
      try
      {
         return CompareOps.isLessThan(new integer(a),
                                      new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lt(BigDecimal a, BigDecimal b)
   {
      try
      {
         return CompareOps.isLessThan(new decimal(a),
                                      new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lt(BigDecimal a, Integer b)
   {
      try
      {
         return CompareOps.isLessThan(new decimal(a),
                                      new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt; <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThan(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lt(Integer a, BigDecimal b)
   {
      try
      {
         return CompareOps.isLessThan(new integer(a),
                                      new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lte(Boolean a, Boolean b)
   {
      try
      {
         return CompareOps.isLessThanOrEqual(new logical(a),
                                             new logical(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lte(String a, String b)
   {
      try
      {
         return CompareOps.isLessThanOrEqual(new character(a),
                                             new character(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lte(Integer a, Integer b)
   {
      try
      {
         return CompareOps.isLessThanOrEqual(new integer(a),
                                             new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lte(BigDecimal a, BigDecimal b)
   {
      try
      {
         return CompareOps.isLessThanOrEqual(new decimal(a),
                                             new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lte(BigDecimal a, Integer b)
   {
      try
      {
         return CompareOps.isLessThanOrEqual(new decimal(a),
                                             new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Test whether <code>a</code> &lt;= <code>b</code>.
    *
    * @see  com.goldencode.p2j.util.CompareOps#isLessThanOrEqual(
    *          com.goldencode.p2j.util.BaseDataType,
    *          com.goldencode.p2j.util.BaseDataType)
    */
   public static Boolean lte(Integer a, BigDecimal b)
   {
      try
      {
         return CompareOps.isLessThanOrEqual(new integer(a),
                                             new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Add two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#plus(
    *          com.goldencode.p2j.util.integer,
    *          com.goldencode.p2j.util.integer)
    */
   public static Integer plus(Integer a, Integer b)
   {
      try
      {
         return MathOps.plus(new integer(a), new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Add two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#plus(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal plus(BigDecimal a, BigDecimal b)
   {
      try
      {
         return MathOps.plus(new decimal(a), new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Add two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#plus(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal plus(BigDecimal a, Integer b)
   {
      try
      {
         return MathOps.plus(new decimal(a), new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Add two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#plus(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal plus(Integer a, BigDecimal b)
   {
      try
      {
         return MathOps.plus(new integer(a), new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Add a number of days to a date.
    *
    * @see  com.goldencode.p2j.util.date#plusDays(
    *          com.goldencode.p2j.util.date,
    *          com.goldencode.p2j.util.integer)
    */
   public static Date plus(Date value, Integer days)
   {
      try
      {
         java.util.Date d = date.plusDays(new date(value),
                                          new integer(days)).dateValue();
         
         return (d != null ? new Date(d.getTime()) : null);
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Add a number of days to a date.
    *
    * @see  com.goldencode.p2j.util.date#plusDays(
    *          com.goldencode.p2j.util.date,
    *          com.goldencode.p2j.util.decimal)
    */
   public static Date plus(Date value, BigDecimal days)
   {
      try
      {
         java.util.Date d = date.plusDays(new date(value),
                                          new decimal(days)).dateValue();
         
         return (d != null ? new Date(d.getTime()) : null);
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Add a number of days to a date.
    *
    * @see  com.goldencode.p2j.util.date#plusDays(
    *          com.goldencode.p2j.util.date,
    *          com.goldencode.p2j.util.integer)
    */
   public static Date plus(Integer days, Date value)
   {
      try
      {
         java.util.Date d = date.plusDays(new date(value),
                                          new integer(days)).dateValue();
         
         return (d != null ? new Date(d.getTime()) : null);
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Add a number of days to a date.
    *
    * @see  com.goldencode.p2j.util.date#plusDays(
    *          com.goldencode.p2j.util.date,
    *          com.goldencode.p2j.util.decimal)
    */
   public static Date plus(BigDecimal days, Date value)
   {
      try
      {
         java.util.Date d = date.plusDays(new date(value),
                                          new decimal(days)).dateValue();
         
         return (d != null ? new Date(d.getTime()) : null);
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Subtract one number from another.
    *
    * @see  com.goldencode.p2j.util.MathOps#minus(
    *          com.goldencode.p2j.util.integer,
    *          com.goldencode.p2j.util.integer)
    */
   public static Integer minus(Integer a, Integer b)
   {
      try
      {
         return MathOps.minus(new integer(a), new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Subtract one number from another.
    *
    * @see  com.goldencode.p2j.util.MathOps#minus(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal minus(BigDecimal a, BigDecimal b)
   {
      try
      {
         return MathOps.minus(new decimal(a), new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Subtract one number from another.
    *
    * @see  com.goldencode.p2j.util.MathOps#minus(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal minus(BigDecimal a, integer b)
   {
      try
      {
         return MathOps.minus(new decimal(a), new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Subtract one number from another.
    *
    * @see  com.goldencode.p2j.util.MathOps#minus(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal minus(Integer a, BigDecimal b)
   {
      try
      {
         return MathOps.minus(new integer(a), new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate a date by subtracting a number of days from a date.
    *
    * @see  com.goldencode.p2j.util.date#minusDays(
    *          com.goldencode.p2j.util.date,
    *          com.goldencode.p2j.util.integer)
    */
   public static Date minus(Date value, Integer days)
   {
      try
      {
         java.util.Date d = date.minusDays(new date(value),
                                           new integer(days)).dateValue();
         
         return (d != null ? new Date(d.getTime()) : null);
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate a date by subtracting a number of days from a date.
    *
    * @see  com.goldencode.p2j.util.date#minusDays(
    *          com.goldencode.p2j.util.date,
    *          com.goldencode.p2j.util.decimal)
    */
   public static Date minus(Date value, BigDecimal days)
   {
      try
      {
         java.util.Date d = date.minusDays(new date(value),
                                           new decimal(days)).dateValue();
         
         return (d != null ? new Date(d.getTime()) : null);
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Determine the number of days between two dates.
    *
    * @see  com.goldencode.p2j.util.date#differenceNum(
    *          com.goldencode.p2j.util.date,
    *          com.goldencode.p2j.util.date)
    */
   public static Integer dateSpan(Date d1, Date d2)
   {
      try
      {
         return date.differenceNum(new date(d1), new date(d2)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the product of two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#multiply(
    *          com.goldencode.p2j.util.integer,
    *          com.goldencode.p2j.util.integer)
    */
   public static Integer multiply(Integer a, Integer b)
   {
      try
      {
         return MathOps.multiply(new integer(a), new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the product of two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#multiply(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal multiply(BigDecimal a, BigDecimal b)
   {
      try
      {
         return MathOps.multiply(new decimal(a), new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the product of two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#multiply(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal multiply(BigDecimal a, Integer b)
   {
      try
      {
         return MathOps.multiply(new decimal(a), new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the product of two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#multiply(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal multiply(Integer a, BigDecimal b)
   {
      try
      {
         return MathOps.multiply(new integer(a), new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the quotient of two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#divide(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal divide(Integer a, Integer b)
   {
      try
      {
         return MathOps.divide(new integer(a), new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the quotient of two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#divide(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal divide(BigDecimal a, BigDecimal b)
   {
      try
      {
         return MathOps.divide(new decimal(a), new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the quotient of two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#divide(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal divide(BigDecimal a, Integer b)
   {
      try
      {
         return MathOps.divide(new decimal(a), new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the quotient of two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#divide(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static BigDecimal divide(Integer a, BigDecimal b)
   {
      try
      {
         return MathOps.divide(new integer(a), new decimal(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the integer remainder resulting from the integer division of
    * two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#modulo(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static Integer modulo(Integer a, Integer b)
   {
      try
      {
         return MathOps.modulo(new integer(a), new integer(b)).toJavaType();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the integer remainder resulting from the integer division of
    * two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#modulo(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static Integer modulo(BigDecimal a, BigDecimal b)
   {
      try
      {
         int64 mod = MathOps.modulo(new decimal(a), new decimal(b));
         return mod.isUnknown() ? null : mod.intValue();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the integer remainder resulting from the integer division of
    * two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#modulo(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static Integer modulo(BigDecimal a, Integer b)
   {
      try
      {
         int64 mod = MathOps.modulo(new decimal(a), new integer(b));
         return mod.isUnknown() ? null : mod.intValue();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
   
   /**
    * Calculate the integer remainder resulting from the integer division of
    * two numbers.
    *
    * @see  com.goldencode.p2j.util.MathOps#modulo(
    *          com.goldencode.p2j.util.NumberType,
    *          com.goldencode.p2j.util.NumberType)
    */
   public static Integer modulo(Integer a, BigDecimal b)
   {
      try
      {
         int64 mod = MathOps.modulo(new integer(a), new decimal(b));
         return mod.isUnknown() ? null : mod.intValue();
      }
      catch (RuntimeException exc)
      {
         ErrorHandler.handleError(exc);
         
         return null;
      }
   }
}
